from django.db.migrations import RunPython as RunPythonBase, RunSQL as RunSQLBase


class Variable:
    """
    Wrapper type to be able to format the variable name correctly inside a migration
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __bool__(self):
        return bool(self.value)


class RunPython(RunPythonBase):
    # Fake the class so the OperationWriter thinks its the internal class and not a custom one
    __class__ = RunPythonBase

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()
        kwargs['elidable'] = self.elidable
        return name, args, kwargs

    @classmethod
    def from_operation(cls, operation, unique_names):
        operation.code.__original_qualname__ = operation.code.__qualname__
        operation.code.__qualname__ = unique_names(operation.code.__qualname__)
        if operation.reverse_code:
            operation.reverse_code.__original_qualname__ = operation.reverse_code.__qualname__
            operation.reverse_code.__qualname__ = unique_names(operation.reverse_code.__qualname__)
        return cls(code=operation.code, reverse_code=operation.reverse_code, atomic=operation.atomic,
                   hints=operation.hints, elidable=operation.elidable)


class RunSQL(RunSQLBase):
    # Fake the class so the OperationWriter thinks its the internal class and not a custom one
    __class__ = RunSQLBase

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()
        kwargs['elidable'] = self.elidable
        return name, args, kwargs

    @classmethod
    def from_operation(cls, operation, unique_names):
        name = unique_names('SQL', force_number=True)
        reverse_sql = Variable('%s_ROLLBACK' % name, operation.reverse_sql) if operation.reverse_sql else None

        return cls(sql=Variable(name, operation.sql), reverse_sql=reverse_sql,
                   state_operations=operation.state_operations, hints=operation.hints,
                   elidable=operation.elidable)



from django.conf import settings as global_settings

DJANGO_SQUASH_IGNORE_APPS = getattr(global_settings, 'DJANGO_SQUASH_IGNORE_APPS', None) or []
DJANGO_SQUASH_MIGRATION_NAME = getattr(global_settings, 'DJANGO_SQUASH_MIGRATION_NAME', None) or 'squashed'



import itertools
import os

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError, no_translations
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.state import ProjectState

from django_squash import settings

from .lib.autodetector import SquashMigrationAutodetector
from .lib.loader import SquashMigrationLoader
from .lib.questioner import NonInteractiveMigrationQuestioner
from .lib.writer import MigrationWriter


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--only', action='append', nargs='*', help='Only squash the specified apps'
        )
        parser.add_argument(
            '--ignore-app',  action='append', nargs='*', default=settings.DJANGO_SQUASH_IGNORE_APPS,
            help='Ignore app name from quashing, ensure that there is nothing dependent on these apps. '
                 '(default: %(default)s)',
        )
        parser.add_argument(
            '--dry-run', action='store_true', dest='dry_run',
            help="Just show what migrations would be made; don't actually write them.",
        )
        parser.add_argument(
            '--squashed-name', default=settings.DJANGO_SQUASH_MIGRATION_NAME,
            help='Sets the name of the new squashed migration. Also accepted are the standard datetime parse '
                 'variables such as "%%Y%%m%%d". (default: "%(default)s" -> "xxxx_%(default)s")',
        )

    @no_translations
    def handle(self, **kwargs):
        self.verbosity = 1
        self.include_header = False
        self.dry_run = kwargs['dry_run']

        ignore_apps = []
        bad_apps = []

        for app_labels in kwargs['ignore_app']:
            for app_label in app_labels:
                try:
                    apps.get_app_config(app_label)
                    ignore_apps.append(app_label)
                except (LookupError, TypeError):
                    bad_apps.append(str(app_label))

        if kwargs['only']:
            only_apps = []

            for app_labels in kwargs['only']:
                for app_label in app_labels:
                    try:
                        apps.get_app_config(app_label)
                        only_apps.append(app_label)
                    except (LookupError, TypeError):
                        bad_apps.append(str(app_label))

            for app_name in apps.app_configs.keys():
                if app_name not in only_apps:
                    ignore_apps.append(app_name)

        if bad_apps:
            raise CommandError("The following apps are not valid: %s" % (', '.join(bad_apps)))

        questioner = NonInteractiveMigrationQuestioner(specified_apps=None, dry_run=False)

        loader = MigrationLoader(None, ignore_no_migrations=True)
        squash_loader = SquashMigrationLoader(None, ignore_no_migrations=True)

        # Set up autodetector
        autodetector = SquashMigrationAutodetector(
            squash_loader.project_state(),
            ProjectState.from_apps(apps),
            questioner,
        )

        squashed_changes = autodetector.squash(
            real_loader=loader,
            squash_loader=squash_loader,
            ignore_apps=ignore_apps,
            migration_name=kwargs['squashed_name']
        )

        replacing_migrations = 0
        for migration in itertools.chain.from_iterable(squashed_changes.values()):
            replacing_migrations += len(migration.replaces)

        if not replacing_migrations:
            raise CommandError("There are no migrations to squash.")

        self.write_migration_files(squashed_changes)

    def write_migration_files(self, changes):
        """
        Take a changes dict and write them out as migration files.
        """
        directory_created = {}
        for app_label, app_migrations in changes.items():
            if self.verbosity >= 1:
                self.stdout.write(self.style.MIGRATE_HEADING("Migrations for '%s':" % app_label) + "\n")
            for migration in app_migrations:
                # Describe the migration
                writer = MigrationWriter(migration, self.include_header)
                if self.verbosity >= 1:
                    # Display a relative path if it's below the current working
                    # directory, or an absolute path otherwise.
                    try:
                        migration_string = os.path.relpath(writer.path)
                    except ValueError:
                        migration_string = writer.path
                    if migration_string.startswith('..'):
                        migration_string = writer.path
                    self.stdout.write("  %s\n" % (self.style.MIGRATE_LABEL(migration_string),))
                    if hasattr(migration, 'is_migration_level') and migration.is_migration_level:
                        for operation in migration.describe():
                            self.stdout.write("    - %s\n" % operation)
                    else:
                        for operation in migration.operations:
                            self.stdout.write("    - %s\n" % operation.describe())
                if not self.dry_run:
                    # Write the migrations file to the disk.
                    migrations_directory = os.path.dirname(writer.path)
                    if not directory_created.get(app_label):
                        os.makedirs(migrations_directory, exist_ok=True)
                        init_path = os.path.join(migrations_directory, "__init__.py")
                        if not os.path.isfile(init_path):
                            open(init_path, "w").close()
                        # We just do this once per app
                        directory_created[app_label] = True
                    migration_string = writer.as_string()
                    if migration_string is None:
                        # File was deleted
                        continue
                    with open(writer.path, "w", encoding='utf-8') as fh:
                        fh.write(migration_string)
                elif self.verbosity == 3:
                    # Alternatively, makemigrations --dry-run --verbosity 3
                    # will output the migrations to stdout rather than saving
                    # the file to the disk.
                    self.stdout.write(self.style.MIGRATE_HEADING(
                        "Full migrations file '%s':" % writer.filename) + "\n"
                    )
                    self.stdout.write("%s\n" % writer.as_string())



import ast
import inspect
import os
import re

from django import get_version
from django.db import migrations as migration_module
from django.db.migrations.writer import (
    MIGRATION_HEADER_TEMPLATE, MIGRATION_TEMPLATE, MigrationWriter as MigrationWriterBase, OperationWriter,
)
from django.utils.timezone import now


def find_brackets(line, p_count, b_count):
    for char in line:
        if char == '(':
            p_count += 1
        elif char == ')':
            p_count -= 1
        elif char == '[':
            b_count += 1
        elif char == ']':
            b_count -= 1
    return p_count, b_count


def replace_migration_attribute(source, attr, value):
    tree = ast.parse(source)
    # Skip this file if it is not a migration.
    migration_node = None
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == 'Migration':
            migration_node = node
            break
    else:
        return

    # Find the `attr` variable.
    comment_out_nodes = {}
    for node in migration_node.body:
        if isinstance(node, ast.Assign) and node.targets[0].id == attr:
            comment_out_nodes[node.lineno] = (node.targets[0].col_offset, node.targets[0].id,)

    # Skip this migration if it does not contain the `attr` we're looking for
    if not comment_out_nodes:
        return

    # Remove the lines that form the multi-line "replaces" statement.
    p_count = 0
    b_count = 0
    col_offset = None
    name = None
    output = []
    for lineno, line in enumerate(source.splitlines()):
        if lineno + 1 in comment_out_nodes.keys():
            output.append(' ' * comment_out_nodes[lineno + 1][0] + attr + ' = ' + str(value))
            p_count = 0
            b_count = 0
            col_offset, name = comment_out_nodes[lineno + 1]
            p_count, b_count = find_brackets(line, p_count, b_count)
        elif p_count != 0 or b_count != 0:
            p_count, b_count = find_brackets(line, p_count, b_count)
        else:
            output.append(line)

    # Overwrite the existing migration file to update it.
    return '\n'.join(output) + '\n'


class ReplacementMigrationWriter(MigrationWriterBase):
    """
    Take a Migration instance and is able to produce the contents
    of the migration file from it.
    """
    template_class_header = MIGRATION_HEADER_TEMPLATE
    template_class = MIGRATION_TEMPLATE

    def __init__(self, migration, include_header=True):
        self.migration = migration
        self.include_header = include_header
        self.needs_manual_porting = False

    def as_string(self):
        """Return a string of the file contents."""
        return self.template_class % self.get_kwargs()

    def get_kwargs(self):
        items = {
            "replaces_str": "",
            "initial_str": "",
        }

        imports = set()

        # Deconstruct operations
        operations = []
        for operation in self.migration.operations:
            operation_string, operation_imports = OperationWriter(operation).serialize()
            imports.update(operation_imports)
            operations.append(operation_string)
        items["operations"] = "\n".join(operations) + "\n" if operations else ""

        # Format dependencies and write out swappable dependencies right
        dependencies = []
        for dependency in self.migration.dependencies:
            if dependency[0] == "__setting__":
                dependencies.append("        migrations.swappable_dependency(settings.%s)," % dependency[1])
                imports.add("from django.conf import settings")
            else:
                dependencies.append("        %s," % self.serialize(dependency)[0])
        items["dependencies"] = "\n".join(dependencies) + "\n" if dependencies else ""

        # Format imports nicely, swapping imports of functions from migration files
        # for comments
        migration_imports = set()
        for line in list(imports):
            if re.match(r"^import (.*)\.\d+[^\s]*$", line):
                migration_imports.add(line.split("import")[1].strip())
                imports.remove(line)
                self.needs_manual_porting = True

        # django.db.migrations is always used, but models import may not be.
        # If models import exists, merge it with migrations import.
        if "from django.db import models" in imports:
            imports.discard("from django.db import models")
            imports.add("from django.db import migrations, models")
        else:
            imports.add("from django.db import migrations")

        # Sort imports by the package / module to be imported (the part after
        # "from" in "from ... import ..." or after "import" in "import ...").
        sorted_imports = sorted(imports, key=lambda i: i.split()[1])
        items["imports"] = "\n".join(sorted_imports) + "\n" if imports else ""
        if migration_imports:
            items["imports"] += (
                "\n\n# Functions from the following migrations need manual "
                "copying.\n# Move them and any dependencies into this file, "
                "then update the\n# RunPython operations to refer to the local "
                "versions:\n# %s"
            ) % "\n# ".join(sorted(migration_imports))
        # If there's a replaces, make a string for it
        if self.migration.replaces:
            items['replaces_str'] = "\n    replaces = %s\n" % self.serialize(self.migration.replaces)[0]
        # Hinting that goes into comment
        if self.include_header:
            items['migration_header'] = self.template_class_header % {
                'version': get_version(),
                'timestamp': now().strftime("%Y-%m-%d %H:%M"),
            }
        else:
            items['migration_header'] = ""

        if self.migration.initial:
            items['initial_str'] = "\n    initial = True\n"

        return items


class MigrationWriter(ReplacementMigrationWriter):
    template_class = """\
%(migration_header)s%(imports)s%(functions)s%(variables)s

class Migration(migrations.Migration):
%(replaces_str)s%(initial_str)s
    dependencies = [
%(dependencies)s\
    ]

    operations = [
%(operations)s\
    ]
"""

    template_variable = '''%s = """%s"""'''

    def as_string(self):
        if hasattr(self.migration, 'is_migration_level') and self.migration.is_migration_level:
            return self.replace_in_migration()
        else:
            return super().as_string()

    def replace_in_migration(self):
        if self.migration._deleted:
            os.remove(self.path)
            return

        changed = False
        with open(self.path) as f:
            source = f.read()

        if self.migration._dependencies_change:
            source = replace_migration_attribute(source, 'dependencies', self.migration.dependencies)
            changed = True
        if self.migration._replaces_change:
            source = replace_migration_attribute(source, 'replaces', self.migration.replaces)
            changed = True
        if not changed:
            raise NotImplementedError()

        return source

    @staticmethod
    def extract_function(code):
        function_source = inspect.getsource(code)
        if code.__original_qualname__ == code.__qualname__:
            return function_source

        function_source = re.sub(rf'(def\s+){code.__original_qualname__}',
                                 rf'\1{code.__qualname__}',
                                 function_source,
                                 1)
        return function_source

    def get_kwargs(self):
        kwargs = super().get_kwargs()

        functions = []
        variables = []
        for operation in self.migration.operations:
            if isinstance(operation, migration_module.RunPython):
                functions.append(self.extract_function(operation.code))
                if operation.reverse_code:
                    functions.append(self.extract_function(operation.reverse_code))
            elif isinstance(operation, migration_module.RunSQL):
                variables.append(self.template_variable % (operation.sql.name, operation.sql.value))
                if operation.reverse_sql:
                    variables.append(self.template_variable % (operation.reverse_sql.name,
                                     operation.reverse_sql.value))

        kwargs['functions'] = ('\n\n' if functions else '') + '\n\n'.join(functions)
        kwargs['variables'] = ('\n\n' if variables else '') + '\n\n'.join(variables)
        kwargs['operations'] = kwargs['operations'].replace('DELETEMEPLEASE.', '')
        kwargs['imports'] = kwargs['imports'].replace('import DELETEMEPLEASE\n', '')

        imports = (x for x in set(kwargs['imports'].split('\n') + getattr(self.migration, 'extra_imports', [])) if x)
        sorted_imports = sorted(imports, key=lambda i: i.split()[1])
        kwargs["imports"] = "\n".join(sorted_imports) + "\n" if imports else ""

        return kwargs



from django.db.migrations.questioner import NonInteractiveMigrationQuestioner as NonInteractiveMigrationQuestionerBase


class NonInteractiveMigrationQuestioner(NonInteractiveMigrationQuestionerBase):
    def ask_initial(self, *args, **kwargs):
        # Ensures that the 0001_initial will always be generated
        return True



import inspect
import os
import tempfile
from contextlib import ExitStack

from django.apps import apps
from django.conf import settings
from django.db.migrations.loader import MigrationLoader


class SquashMigrationLoader(MigrationLoader):
    def __init__(self, *args, **kwargs):
        original_migration_modules = settings.MIGRATION_MODULES.copy()
        project_path = os.path.abspath(os.curdir)

        with ExitStack() as stack:
            # Find each app that belongs to the user and are not in the site-packages. Create a fake temporary
            # directory inside each app that will tell django we don't have any migrations at all.
            for app_config in apps.get_app_configs():
                module = app_config.module
                app_path = os.path.dirname(os.path.abspath(inspect.getsourcefile(module)))

                if app_path.startswith(project_path):
                    temp_dir = stack.enter_context(tempfile.TemporaryDirectory(prefix='migrations_', dir=app_path))
                    # Need to make this directory a proper python module otherwise django will refuse to recognize it
                    open(os.path.join(temp_dir, '__init__.py'), 'a').close()
                    settings.MIGRATION_MODULES[app_config.label] = '%s.%s' % (app_config.module.__name__,
                                                                              os.path.basename(temp_dir))

            super().__init__(*args, **kwargs)

        settings.MIGRATION_MODULES = original_migration_modules



import ast
import datetime
import inspect
import itertools
import os
import sys
import types
from collections import defaultdict

from django.apps import apps
from django.conf import settings
from django.db import migrations as migration_module
from django.db.migrations.autodetector import MigrationAutodetector as MigrationAutodetectorBase

from .operators import RunPython, RunSQL


def source_directory(module):
    return os.path.dirname(os.path.abspath(inspect.getsourcefile(module)))


class Migration(migration_module.Migration):

    _deleted = False
    _dependencies_change = False
    _replaces_change = False

    def describe(self):
        if self._deleted:
            yield 'Deleted'
        if self._dependencies_change:
            yield '"dependencies" changed'
        if self._replaces_change:
            yield '"replaces" keyword removed'

    @property
    def is_migration_level(self):
        return self._deleted or self._dependencies_change or self._replaces_change

    def __getitem__(self, index):
        return (self.app_label, self.name)[index]

    def __iter__(self):
        yield from (self.app_label, self.name)

    @classmethod
    def from_migration(cls, migration):
        new = Migration(name=migration.name, app_label=migration.app_label)
        new.__dict__ = migration.__dict__.copy()
        new._original_migration = migration
        return new


class UniqueVariableName:
    def __init__(self):
        self.names = defaultdict(int)

    def __call__(self, name, force_number=False):
        self.names[name] += 1
        count = self.names[name]
        if not force_number and count == 1:
            return name
        else:
            new_name = '%s_%s' % (name, count)
            # Make sure that the function name is fully unique
            # You can potentially have the same name already defined.
            return self(new_name)


def all_custom_operations(operations, unique_names):
    """
    Generator that loops over all the operations and traverses sub-operations such as those inside a -
    SeparateDatabaseAndState class.
    """

    for operation in operations:
        if operation.elidable:
            continue

        if isinstance(operation, migration_module.RunSQL):
            yield RunSQL.from_operation(operation, unique_names)
        elif isinstance(operation, migration_module.RunPython):
            yield RunPython.from_operation(operation, unique_names)
        elif isinstance(operation, migration_module.SeparateDatabaseAndState):
            # A valid use case for this should be given before any work is done.
            pass


def get_imports(module):
    """
    Return an generator with all the imports to a particular py file as string
    """
    source = inspect.getsource(module)
    path = inspect.getsourcefile(module)

    root = ast.parse(source, path)
    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.Import):
            for n in node.names:
                yield f'import {n.name}'
        elif isinstance(node, ast.ImportFrom):
            module = node.module.split('.')
            # Remove old python 2.x imports
            if '__future__' not in node.module:
                yield f"from {node.module} import {', '.join([x.name for x in node.names])}"
        else:
            continue


def copy_func(f, name=None):
    func = types.FunctionType(f.__code__, f.__globals__, name or f.__qualname__,
                              f.__defaults__, f.__closure__)
    func.__qualname__ = f.__qualname__
    func.__original_qualname__ = f.__original_qualname__
    return func


class SquashMigrationAutodetector(MigrationAutodetectorBase):

    def add_non_elidables(self, original, loader, changes):
        unique_names = UniqueVariableName()
        replacing_migrations_by_app = {app: [original.disk_migrations[r]
                                             for r in itertools.chain.from_iterable([m.replaces for m in migrations])]
                                       for app, migrations in changes.items()}

        for app in changes.keys():
            operations = []
            imports = []

            for migration in replacing_migrations_by_app[app]:
                module = sys.modules[migration.__module__]
                imports.extend(get_imports(module))
                for operation in all_custom_operations(migration.operations, unique_names):
                    if isinstance(operation, migration_module.RunPython):
                        operation.code = copy_func(operation.code)
                        operation.code.__module__ = 'DELETEMEPLEASE'  # TODO: get a better name?
                        if operation.reverse_code:
                            operation.reverse_code = copy_func(operation.reverse_code)
                            operation.reverse_code.__module__ = 'DELETEMEPLEASE'  # TODO: get a better name?
                    operations.append(operation)

            migration = changes[app][-1]
            migration.operations += operations
            migration.extra_imports = imports

    def replace_current_migrations(self, original, graph, changes):
        """
        Adds 'replaces' to the squash migrations with all the current apps we have.
        """
        migrations_by_app = defaultdict(list)
        for app, migration in original.graph.node_map:
            migrations_by_app[app].append((app, migration))

        for app, migrations in changes.items():
            for migration in migrations:
                # TODO: maybe use a proper order???
                migration.replaces = sorted(migrations_by_app[app])

    def rename_migrations(self, original, graph, changes, migration_name):
        """
        Continues the numbering from whats there now.
        """
        current_counters_by_app = defaultdict(int)
        for app, migration in original.graph.node_map:
            current_counters_by_app[app] = max([int(migration[:4]), current_counters_by_app[app]])

        for app, migrations in changes.items():
            for migration in migrations:
                next_number = current_counters_by_app[app] = current_counters_by_app[app] + 1
                migration_name = datetime.datetime.now().strftime(migration_name)
                migration.name = "%04i_%s" % (
                    next_number,
                    migration_name or 'squashed',
                )

    def convert_migration_references_to_objects(self, original, graph, changes):
        """
        Swap django.db.migrations.Migration with a custom one that behaves like a tuple when read, but is still an
        object for the purpose of easy renames.
        """
        migrations_by_name = {}

        # First pass, swapping new objects
        for app_label, migrations in changes.items():
            new_migrations = []
            for migration in migrations:
                migration_id = migration.app_label, migration.name
                new_migration = Migration.from_migration(migration)
                migrations_by_name[migration_id] = new_migration
                new_migrations.append(new_migration)
            changes[app_label] = new_migrations

        # Second pass, replace the tuples with the newly created objects
        for app_label, migrations in changes.items():
            for migration in migrations:
                new_dependencies = []
                for dependency in migration.dependencies:
                    if dependency[0] == "__setting__":
                        app_label = getattr(settings, dependency[1]).split('.')[0]
                        migrations = [migration for (app, _), migration in migrations_by_name.items()
                                      if app == app_label]
                        dependency = tuple(migrations[-1])
                    elif dependency[1] == "__first__":
                        dependency = original.graph.root_nodes(dependency[0])[0]
                    elif dependency[1] == "__latest__":
                        dependency = original.graph.leaf_nodes(dependency[0])[0]

                    migration_id = dependency
                    if migration_id not in migrations_by_name:
                        new_migration = Migration.from_migration(original.disk_migrations[migration_id])
                        migrations_by_name[migration_id] = new_migration
                    new_dependencies.append(migrations_by_name[migration_id])

                migration.dependencies = new_dependencies

    def create_deleted_models_migrations(self, loader, changes):
        migrations_by_label = defaultdict(list)
        for (app, ident), _ in itertools.groupby(loader.disk_migrations.items(), lambda x: x[0]):
            migrations_by_label[app].append(ident)

        for app_config in loader.project_state().apps.get_app_configs():
            if app_config.models and app_config.label in migrations_by_label:
                migrations_by_label.pop(app_config.label)

        for app_label, migrations in migrations_by_label.items():
            subclass = type("Migration", (Migration,), {"operations": [], "dependencies": []})
            instance = subclass("temp", app_label)
            instance.replaces = migrations
            changes[app_label] = [instance]

    def squash(self, real_loader, squash_loader, ignore_apps=None, migration_name=None):
        changes_ = self.delete_old_squashed(real_loader, ignore_apps)

        graph = squash_loader.graph
        changes = super().changes(graph, trim_to_apps=None, convert_apps=None, migration_name=None)

        for app in ignore_apps:
            changes.pop(app, None)

        self.create_deleted_models_migrations(real_loader, changes)
        self.convert_migration_references_to_objects(real_loader, graph, changes)
        self.rename_migrations(real_loader, graph, changes, migration_name)
        self.replace_current_migrations(real_loader, graph, changes)
        self.add_non_elidables(real_loader, squash_loader, changes)

        for app, change in changes_.items():
            changes[app].extend(change)

        return changes

    def delete_old_squashed(self, loader, ignore_apps=None):
        changes = defaultdict(set)
        project_path = os.path.abspath(os.curdir)
        project_apps = [app.label for app in apps.get_app_configs()
                        if source_directory(app.module).startswith(project_path)]

        real_migrations = (Migration.from_migration(loader.disk_migrations[key])
                           for key in loader.graph.node_map.keys())
        project_migrations = [migration for migration in real_migrations
                              if migration.app_label in project_apps and migration.app_label not in ignore_apps or []]
        replaced_migrations = [Migration.from_migration(migration)
                               for migration in project_migrations if migration.replaces]

        migrations_to_remove = set()
        for migration in (y for x in replaced_migrations for y in x.replaces if y[0] not in ignore_apps or []):
            real_migration = Migration.from_migration(loader.disk_migrations[migration])
            real_migration._deleted = True
            migrations_to_remove.add(migration)
            changes[migration[0]].add(real_migration)

        # Remove all the old dependencies that will be removed
        for migration in project_migrations:
            new_dependencies = [migration for migration in migration.dependencies
                                if migration not in migrations_to_remove]
            if new_dependencies == migration.dependencies:
                # There is no need to write anything
                continue
            migration._dependencies_change = True
            changes[migration.app_label].add(migration)
            setattr(migration, 'dependencies', new_dependencies)

        for migration in replaced_migrations:
            migration._replaces_change = True
            changes[migration.app_label].add(migration)
            setattr(migration, 'replaces', [])

        return changes



#!/usr/bin/env python
# coding=utf-8

import io
import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as fp:
    README = fp.read()


class DjangoTest(TestCommand):
    command_consumes_arguments = True

    def initialize_options(self):
        self.args = None
        super().initialize_options()

    def run_tests(self):
        from tests.runtests import django_tests

        sys.path.insert(0, 'tests')
        os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')

        django_tests(verbosity=3,
                     interactive=False,
                     failfast=True,
                     keepdb=False,
                     reverse=False,
                     test_labels=[os.path.normpath(labels) for labels in self.args],
                     debug_sql=False,
                     parallel=False,
                     tags=[],
                     exclude_tags=[],
                     test_name_patterns=[])


setup(
    name='django_squash',
    version='0.0.8',
    description="A migration squasher that doesn't care how Humpty Dumpty was put together.",
    long_description=README,
    classifiers=[
        # See https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='django migration squashing squash',
    author='Javier Buzzi',
    author_email='buzzi.javier@gmail.com',
    url='https://github.com/kingbuzzman/django-squash',
    license="MIT",
    packages=find_packages(exclude=['tests*']),
    platforms=['any'],
    zip_safe=True,
    python_requires='>=3.7',
    cmdclass={'test': DjangoTest},
    install_requires=[
        'django>=3.2',
    ],
    tests_require=[],
    extras_require={
        'test':  ['isort', 'flake8', 'ipdb', 'coverage'],
    }
)



from django.apps import AppConfig
from django.db.migrations.serializer import Serializer


class DjangoSquashConfig(AppConfig):
    name = 'django_squash'

    def ready(self):
        from .management.commands.lib.operators import Variable
        from .management.commands.lib.serializer import VariableSerializer

        Serializer.register(Variable, VariableSerializer)
